/*-------------------------------------------------------------
 *  Copyright (c) 2010, MSS ComBlock. All rights reserved.
 *
 * 	Filename: fpga.c
 *  Last modified: Jun 26, 2010
 *  Authors:
 *  	Alain Zarembowitch / MSS ComBlock
 *  Version: 1.0
 *  Inheritance: N/A
 *  Description: Utilities to configure the FPGA.
 *  IMPORTANT ASSUMPTION: data file is BIT SWAPPED (select BIT SWAP when generating the bit/mcs file in Xilinx ISE)
 *
 *  Dependencies/Pre-requisite: N/A
 -------------------------------------------------------------*/
#include "main.h"
#include "fpga.h"
#include "nandFlash.h"

/*-------- change this section according to the schematics ----------------------------------------*/
inline long fpga_check_initb(void){
	return FIO0PIN & 0x00000080;	// pin P0[7] INIT#
}

inline long fpga_check_done(void){
	return FIO0PIN & 0x00000040;	// pin P0[6] DONE
}

inline void initb2web(void){
	// after configuration change init# input pin to we# output
	FIO0DIR |= 0x00000080;	// p0[7] is now an output
	PIN_FPGA_WEB_H();
	return;
}

// FPGA/NAND Flash selections are mutually exclusive (they share the same data bus)
// enable/disable functions are thus joined at the hip.
inline void FPGA_NFn(void){
	PIN_NAND_CEB_H();	// make sure NAND flash is deselected, as NF and FPGA share control signals
	PIN_FPGA_CSIB_L();	// enable FPGA
}
inline void NF_FPGAn(void){
	PIN_FPGA_CSIB_H();	// disable FPGA
	PIN_NAND_CEB_L();	// enable NAND flash
}

/*-------- end of schematics-dependent code ----------------------------------------*/


long fpga_config(unsigned long start_addr){
	// configure the FPGA with the configuration data stored in NAND flash.
	// config_no ranges from 0 to NSWCONFIGS-1
	unsigned long nblocks, start_block, iblock, iblock_alt, ipage;
	unsigned long rptr, end_addr;
	unsigned long i;
	unsigned char pageBuf[PAGESIZE];	// page from the NAND memory
	long retval;

	// rules for FPGA configuration storage in NAND flash:
	// 1. each fpga configuration has the exclusive use of N consecutive 128KB blocks
	// 2. the fpga configuration file starts at the first column of the first page of the first block
	// 3. the flash may re-direct a block address to the upper end if a block is flagged as bad.

	// checks
	if(start_addr == 0xFFFFFFFF)
		return -FPGA_OUTOFRANGE;
	if(start_addr > NBLOCKS*BLOCKSIZE)
		return -FPGA_OUTOFRANGE;

	nblocks = (FPGA_CONFIG_SIZE / BLOCKSIZE) + 1;	// rounding up the number of blocks required to store a complete FPGA configuration

	// NAND start/end addresses
	rptr = start_addr;
	start_block = start_addr / BLOCKSIZE;
	end_addr = start_addr + FPGA_CONFIG_SIZE;

	// The key waveforms are specified in Xilinx UG380 Figure 2-9

	// prepare FPGA for programming
	PIN_NAND_CEB_H();	// make sure NAND flash is deselected, as NF and FPGA share control signals
	PIN_FPGA_PROGB_H();
	PIN_FPGA_RWB_H();
	PIN_FPGA_CSIB_H();
	PIN_FPGA_CCLK_H();

	// initialize FPGA
	PIN_FPGA_PROGB_L();
	PIN_FPGA_RWB_L();
	PIN_FPGA_CCLK_L();	// needed for sure (for timing?)
	PIN_FPGA_CCLK_H();
	PIN_FPGA_PROGB_H();

	// await INIT# going high (measured 600us)
	timer[0] = 2; 	// 10ms units -> 10-20ms timeout
	while(1){
		if(timer[0] == 0){	// timer expired
			return -FPGA_ETIMEOUT; // timeout
		}
		if(fpga_check_initb())
			break;
	}

	// read one page at a time
	for(iblock = start_block; rptr <= end_addr; iblock++){
		// 64 pages per block
		// redirect if bad block
		if(iblock >= sizeof(nfInfo.blockRedirect)){
			return -FPGA_OUTOFRANGE;
		}
		iblock_alt = nfInfo.blockRedirect[iblock];
		for(ipage = 0; ipage < 64; ipage++){
			// read a 2KB page from the NAND flash memory
			/* before timing optimization
			PIN_FPGA_CSIB_H();	// disable FPGA
			PIN_NAND_ALE_L();
			PIN_NAND_CEB_L();	// enable NAND flash
			retval = nandFlash_read(rptr, &pageBuf[0], sizeof(pageBuf));
			PIN_NAND_CEB_H();	// disable NAND flash
			*/
			/* After timing optimization */
			PIN_FPGA_CSIB_H();	// disable FPGA
			PIN_NAND_ALE_L();
			PIN_NAND_CEB_L();	// enable NAND flash
			retval = nandFlash_page_read(iblock_alt, ipage, 0, sizeof(pageBuf), &pageBuf[0]);
			PIN_NAND_CEB_H();	// disable NAND flash
			/* end of timing optimization */

			if(retval)
				return retval;
			rptr += PAGESIZE;

			// copy 2KB page to FPGA
			//FIO1CLR = 0x00000002;	// TEST TEST TEST
			busDirOut();
			PIN_FPGA_RWB_L();	// make sure that RWB is always in the write position when CSIB is enabled (otherwise it means abort)
			PIN_FPGA_CCLK_H();
			PIN_FPGA_CSIB_L();	// enable FPGA
			/* before timing optimization
			for(i = 0; i < sizeof(pageBuf); i++){
				busDataOut(pageBuf[i]);	// IMPORTANT ASSUMPTION: data file is BIT SWAPPED (select BIT SWAP when generating the bit/mcs file in Xilinx ISE)
				PIN_FPGA_CCLK_L();
				PIN_FPGA_CCLK_H();
			}
			*/
			/* after timing optimization */
			for(i = 0; i < sizeof(pageBuf); i++){
				FIO2PIN = pageBuf[i] & 0x000001FF;
				PIN_FPGA_CCLK_H();
			}
			/* end of timing optimization */
			PIN_FPGA_CSIB_H();

			if(rptr > end_addr)
				break;	// done. no need to continue to the end of the block.
		}
	}

	// disable all FPGA control lines
	PIN_FPGA_CSIB_H();
	PIN_FPGA_RWB_H();
	// read back (debug only, we need the SelectMAP pins for I/O during operations).
	// fpga_readback(); // untested
	initb2web();	// init# input is now we# output
	return FPGA_OK;
}


long fpga_readback(void){
	// readback FPGA configuration.
	// Three key requirements: BitGen security setting must not prohibit readback (-g Security:none), and bitstream
	//  encryption must not be used. Additionally, the port must be set to retain its function after configuration by
	// setting the persist option in BitGen (-g Persist:Yes), otherwise the SelectMAP data pins
	// revert to user I/O, precluding further configuration operations.
	// See Xilinx ISE generate programming file | properties GUI
	unsigned long nblocks, iblock, ipage, icol;
	unsigned long i;
	unsigned char pageBuf[2048];	// page from the NAND memory

	// waveform as per ug380 Figure 6-1
	PIN_FPGA_CSIB_H();
	PIN_FPGA_RWB_H();
	PIN_FPGA_CCLK_L();
	PIN_FPGA_CCLK_H();
	busDirIn();
	// read one page at a time
	nblocks = (FPGA_CONFIG_SIZE / BLOCKSIZE) + 1;	// rounding up the number of blocks required to store a complete FPGA configuration
	for(iblock = 0; iblock < nblocks; iblock++){
		// 64 pages per block
		for(ipage = 0; ipage < 64; ipage++){
			for(icol = 0; icol < 2048; i++){
				PIN_FPGA_CCLK_L();
				PIN_FPGA_CCLK_H();
				pageBuf[i] = busDataIn();
			}
		}
	}
	PIN_FPGA_CSIB_H();
	return 0;
}

long block_to_config(unsigned long blockNo){
	// returns the FPGA configuration number (starting at 0) based on the NAND flash block number
	long configno;
	configno = blockNo / NFBLOCKSPERCONFIG;
	return configno;
}

void fpga_reset(void){
	// erase FPGA configuration
	PIN_FPGA_PROGB_L();
	PIN_FPGA_RWB_L();
	PIN_FPGA_CCLK_L();	// needed for sure (for timing?)
	PIN_FPGA_CCLK_H();
	PIN_FPGA_PROGB_H();
	return;
}

unsigned char fpga_getByte(unsigned long addr){
	// read byte from the specified FPGA address
	// The number of addresses supported depend on the FPGA code.
	unsigned char a;

	PIN_FPGA_REB_H();	// read#
	PIN_FPGA_WEB_H();	// write#
	PIN_FPGA_ALE_L();	// address latch enable
	busDirOut();
	FPGA_NFn();	// enable FPGA

	busDataOut((unsigned char)(addr));
	PIN_FPGA_ALE_H();	// address latch enable
	PIN_FPGA_WEB_L();	// write#
	PIN_FPGA_WEB_H();	// write#
	PIN_FPGA_ALE_L();	// address latch enable

	busDirIn();
	PIN_FPGA_REB_L();	// read#
	a = busDataIn();
	PIN_FPGA_REB_H();	// read#
	PIN_FPGA_CSIB_H();	// deselect FPGA
	return a;
}

void fpga_setByte(unsigned long addr, unsigned char a){
	// write a byte to the specified FPGA address
	// The number of addresses supported depend on the FPGA code.

	PIN_FPGA_REB_H();	// read#
	PIN_FPGA_WEB_H();	// write#
	PIN_FPGA_ALE_L();	// address latch enable
	busDirOut();
	FPGA_NFn();	// enable FPGA

	busDataOut((unsigned char)(addr));
	PIN_FPGA_ALE_H();	// address latch enable
	PIN_FPGA_WEB_L();	// write#
	PIN_FPGA_WEB_H();	// write#
	PIN_FPGA_ALE_L();	// address latch enable

	busDataOut(a);
	PIN_FPGA_WEB_L();	// write#
	PIN_FPGA_WEB_H();	// write#
	PIN_FPGA_CSIB_H();	// deselect FPGA
	return;
}

long fpga_config_exists(unsigned long start_addr){
	// check the first few bytes in NAND flash to verify that the FPGA configuration exists.
	// otherwise, there one could damage the FPGA by programming all FFs.
	unsigned char pageBuf[PAGESIZE];	// page from the NAND memory
	long retval, i;

	retval = nandFlash_read(start_addr, &pageBuf[0], sizeof(pageBuf));
	for(i = 0; i<sizeof(pageBuf); i++){
		if(pageBuf[i] != 0xFF)
			return 1;	// non-FF character present
	}
	return 0;	// blank page.
	// It is most likely that fpga configuration is not programmed into the NAND flash memory,
	// at least at this address.


}
