/*-------------------------------------------------------------
 *  Copyright (c) 2010-2013, MSS ComBlock. All rights reserved.
 *
 * 	Filename: nandFlash.c
 *  Last modified: 4/17/14
 *  Authors:
 *  	Alain Zarembowitch / MSS ComBlock
 *  Version: 2.0
 *  Inheritance: N/A
 *  Description: NAND Flash driver
 *  TODO Add NOPs as needed to meet the flash timing requirements
 *  see Micron 1Gb NAND flash memory MT29F1Gxx specifications
 *  This driver does not support all possible NAND flash commands. The following commands
 *  are included: reset, page read, read status, readID, block erase, page program.
 *
 *  Dependencies:
 *  1. one must call nandFlash_init() once at power-up/reset before any other NAND flash operation can take place.
 *  2. NAND Flash Ready/Busy# signal must be pulled-up in InitConfig.c or using a PU resistor.
 *  3. This code is written specifically for a 1Gb x8-bit wide NAND Flash (will not work for x16)
 *
 *  Rev 2 2/26/12 AZ
 *  Added define statement to allow for 2,4 and 8Gb NAND flash ICs
 *
 *  Rev 3 4/17/14 AZ
 *  Added support for Spansion S34ML01G in addition to Micron MT29F1G
 -------------------------------------------------------------*/
#include "main.h"
#include "nandFlash.h"
#include "fpga.h"
#include "utils.h"
#include <math.h>

struct nandFlash_info nfInfo;


/*-------- change this section according to the schematics ----------------------------------------*/
inline void busDirOut(void){
	// set 8-bit data bus direction to 'out'
	FIO2DIR |= 0x000003FF;	// UC_DATAx + UC_CTRL as outputs
}

inline void busDirIn(void){
	// set 8-bit data bus direction to 'in'
	FIO2DIR &= ~0x000000FF;	// UC_DATAx as inputs
}

inline unsigned char busDataIn(void){
	return (unsigned char)(FIO2PIN);
}

inline void busDataOut(unsigned char dout){
	FIO2PIN = (FIO2PIN & 0xFFFFFF00) | (((long)(dout)) & 0x000000FF);
}

/*-------- end of schematics-dependent code ----------------------------------------*/
long nandFlash_init(void){
	// initialization. must be called first at power-up/reset.
	long retval, i;

	// 1. call reset the flash memory. CE#=0
	retval = nandFlash_reset();
	if(retval)
		return retval;

	// 2. read Flash information
	nandFlash_readID();

	// 3. survey all bad blocks and build a block re-direction table
	nfInfo.nBadBlocks = 0;
	for(i = 0; i<NBLOCKS; i++){
		if(i >= (NBLOCKS - nfInfo.nBadBlocks))
			break;
		retval = nandFlash_checkBadBlock(i);  // read the first byte in the redundancy section of the first and second page.
		if(retval == NF_OK){
			// block looks good. No re-direction needed
			nfInfo.blockRedirect[i] = i;
		} else if(retval == -NF_EBADBLOCK){
			nfInfo.nBadBlocks++;
			// redirect to a high-numbered block (counting downwards from NBLOCKS)
			nfInfo.blockRedirect[i] = NBLOCKS - nfInfo.nBadBlocks;
		} else {
			return retval;
		}
	}
	// Guaranteed 1,004 (NVB) valid blocks	out of every 1024 total available blocks (if not report error) 1Gb NAND flash
	// Guaranteed 2,008 (NVB) valid blocks	out of every 2,048 total available blocks (if not report error) 2Gb NAND flash
	if(nfInfo.nBadBlocks > 40)
		return -NF_ETOOMANYBADBLOCKS;

	return NF_OK;


}

unsigned char nandFlash_check_status(void){
	// return status byte
	// see Micron 1Gb NAND flash memory MT29F1Gxx specifications Figures 18/48
	// Pre-requisite: CE# must be enabled prior
	unsigned char a;

	busDirOut();
	busDataOut(0x70);	// read status command 0x70
	PIN_NAND_CLE_H();
	PIN_NAND_WEB_L();
	PIN_NAND_WEB_H();
	PIN_NAND_CLE_L();
	busDirIn();
	PIN_NAND_REB_L();
	PIN_NAND_REB_L();	// delay before reading (tREA)
	a = busDataIn();
	PIN_NAND_REB_H();
	return a;	// ready/busy# on bit 5
}

unsigned char nandFlash_check_status_contd(void){
	// follow-on status checks
	// Pre-requisite: call to nandFlash_check_status() first and no other commands in-between)
	// see Micron 1Gb NAND flash memory MT29F1Gxx specifications Figures 18/48
	unsigned char a;
	PIN_NAND_REB_L();
	PIN_NAND_REB_L();	// delay before reading (tREA)
	a = busDataIn();
	PIN_NAND_REB_H();
	return a;
}

long nandFlash_reset(void){
	// returns return 0 on success, a negative error number otherwise.
	// see Micron 1Gb NAND flash memory MT29F1Gxx specifications Figure 37
	unsigned char a;

	// initialize all key signals in idle mode
	busDirOut();
	PIN_NAND_CEB_H();	// CE# = '1'
	PIN_NAND_CLE_L();	// CLE = '0'
	PIN_NAND_ALE_L();	// ALE = '0'
	PIN_NAND_WEB_H();	// WE# = '1'
	PIN_NAND_REB_H();	// RE# = '1'
	PIN_NAND_WPB_L();	// WP# = '0'	write protected

	// reset transaction
	NF_FPGAn();	// enable NAND Flash
	busDataOut(0xFF);	// reset command 0xFF
	PIN_NAND_CLE_H();
	PIN_NAND_WEB_L();
	PIN_NAND_WEB_H();
	PIN_NAND_CLE_L();

	// We must wait for the end of the busy period.
	// set timeout awaiting for ready signal from NAND flash
	timer[0] = 2; 	// 10ms units -> 10-20ms timeout
	while(1){
		if(timer[0] == 0){	// timer expired
			return -NF_ETIMEOUT; // timeout
		}
		a = nandFlash_check_status();
		if((a & 0x60) == 0x60) 	// ready/busy# on bit 5
			break;
	}
	return NF_OK;	// status is successful
}



long nandFlash_readID(void){
	// returns return 0 on success, a negative error number otherwise.
	// see Micron 1Gb NAND flash memory MT29F1Gxx specifications  Figure 15
	// Pre-requisite: CE# must be enabled prior
	int i;
	unsigned char a;

	// check busy status before most commands
	timer[0] = 2; 	// 10ms units -> 10-20ms timeout
	a = nandFlash_check_status();	// first read status
	while(1){
		if((a & 0x60) == 0x60)
			break;
		if(!timer[0])
			return -NF_ETIMEOUT; // timeout
		a = nandFlash_check_status_contd(); // subsequent read status
	}

	busDirOut();
	busDataOut(0x90);	// read ID command
	PIN_NAND_CLE_H();
	PIN_NAND_WEB_L();
	PIN_NAND_WEB_H();
	PIN_NAND_CLE_L();
	PIN_NAND_ALE_H();
	busDataOut(0x00);	// ALE address 0
	PIN_NAND_WEB_L();
	PIN_NAND_WEB_H();
	PIN_NAND_ALE_L();
	busDirIn();
	for(i = 0; i<4; i++){
		PIN_NAND_REB_L();
		PIN_NAND_REB_L();	// delay before reading (tREA)
		a = busDataIn();
		PIN_NAND_REB_H();
		switch (i) {
			case 0:
				nfInfo.manufacturerID = a;	// 0x2C = 44 for Micron
				break;
			case 1:
				nfInfo.deviceID = a;		// 0xD1/0xF1 (TBC)
				break;
			case 3:
				nfInfo.byte3 = a;			// 0x95 = -15
				break;
			default:
				break;
		}
	}
	return NF_OK;
}

long nandFlash_blockErase(unsigned long blockNo){
	// returns return 0 on success, a negative error number otherwise.
	// see Micron 1Gb NAND flash memory MT29F1Gxx specifications Figure 24
	unsigned char a;
	unsigned long addrL;
	unsigned long addrH;	// address spans 4(1Gb) or 5(2Gb+) bytes, thus exceeding a long

	// chip select
	PIN_NAND_ALE_L();	// ALE = '0'
	NF_FPGAn();	// enable NAND Flash

	if(blockNo >= NBLOCKS)
		return -NF_EDATAIN;

	// check busy status before most commands
	timer[0] = 2; 	// 10ms units -> 10-20ms timeout
	a = nandFlash_check_status();	// first read status
	while(1){
		if(a == 0x60)
			break;
		if(!timer[0])
			return -NF_ETIMEOUT; // timeout
		a = nandFlash_check_status_contd(); // subsequent read status
	}

	PIN_NAND_WPB_H();		// remove write protection
	busDirOut();
	busDataOut(0x60);	// Erase setup command
	PIN_NAND_CLE_H();
	PIN_NAND_WEB_L();
	PIN_NAND_WEB_H();
	PIN_NAND_CLE_L();

	addrH = (blockNo >> 10) & 0x00000003;	// used for 2Gb and above
	addrL = (blockNo << 22) & 0xFFC00000;
	//addrL |= (pageNo << 16) & 0x003F0000;
	//addrL |= startCol & 0x0000FFFF;

	PIN_NAND_ALE_H();
	busDataOut((addrL >> 16) & 0x000000FF);
	PIN_NAND_WEB_L();
	PIN_NAND_WEB_H();
	busDataOut((addrL >> 24) & 0x000000FF);
	PIN_NAND_WEB_L();
	PIN_NAND_WEB_H();
	if(MT29F2_4_8G){
		// used for 2Gb and above */
		busDataOut(addrH & 0x000000FF);
		PIN_NAND_WEB_L();
		PIN_NAND_WEB_H();
	}
	PIN_NAND_ALE_L();

	busDataOut(0xD0);	// erase confirm command
	PIN_NAND_CLE_H();
	PIN_NAND_WEB_L();
	PIN_NAND_WEB_H();
	PIN_NAND_CLE_L();

	// NAND flash starts erasing. We must wait for the end of the erase period.
	// set timeout awaiting for ready signal from NAND flash
	// wait until NAND flash is no longer busy  tBERS < 3ms max
	timer[0] = 2; 	// 10ms units -> 10-20ms
	a = nandFlash_check_status();	// first read status
	while(1){
		if((a & 0x40) == 0x40)
			break;	// ready
		if(!timer[0]){
			PIN_NAND_WPB_L();	// re-instate write protection
			PIN_NAND_CEB_H();	// de-select flash
			return -NF_ETIMEOUT; // timeout
		}
		a = nandFlash_check_status_contd(); // subsequent read status
	}

	PIN_NAND_WPB_L();	// reinstate write protection
	PIN_NAND_CEB_H();	// de-select flash
	if(!(a & 0x01))
		return NF_OK;	// successful erase
	else
		return -NF_EBADBLOCK;	// unsuccessful erase
	return NF_OK;
}

// programming is done in 2 or 3 steps: start + optional contd + end.
// part 1.
long nandFlash_page_program_start(unsigned long blockNo, unsigned long pageNo, unsigned long startCol,
			unsigned long nBytes, char *buf){
	// USAGE: PAGES MUST BE PROGRAMMED CONSECUTIVELY WITHIN A BLOCK + BLOCK MUST BE ERASED BEFORE RE-PROGRAMMING
	// Write a full page or part thereof, starting at the specified startCol. The number of bytes to be written is nBytes.
	// The data to be written is stored in buf. The maximum number of bytes to be written is 2112 (for a complete page including the ECC redundancy).
	// returns return 0 on success, a negative error number otherwise.
	// see Micron 1Gb NAND flash memory MT29F1Gxx specifications Figure 19
	unsigned char a;
	unsigned long addrL;
	unsigned long addrH;	// address spans 4(1Gb) or 5(2Gb+) bytes, thus exceeding a long
	long i;

	if(blockNo >= NBLOCKS)
		return -NF_EDATAIN;

	PIN_NAND_ALE_L();	// ALE = '0'
	NF_FPGAn();	// enable NAND Flash

	// check busy status before most commands
	timer[0] = 2; 	// 10ms units -> 10-20ms timeout
	a = nandFlash_check_status();	// first read status
	while(1){
		if((a & 0x60) == 0x60)
			break;
		if(!timer[0])
			return -NF_EBUSY;
		a = nandFlash_check_status_contd(); // subsequent read status
	}

	PIN_NAND_WPB_H();		// remove write protection
	busDirOut();
	busDataOut(0x80);		//page program command
	PIN_NAND_CLE_H();
	PIN_NAND_WEB_L();
	PIN_NAND_WEB_H();
	PIN_NAND_CLE_L();

	// form 5-cycle address array as per Table 4.
	addrH = (blockNo >> 10) & 0x00000003;	// used for 2Gb and above
	addrL = (blockNo << 22) & 0xFFC00000;
	addrL |= (pageNo << 16) & 0x003F0000;
	addrL |= startCol & 0x0000FFFF;

	PIN_NAND_ALE_H();
	busDataOut(addrL & 0x000000FF);
	PIN_NAND_WEB_L();
	PIN_NAND_WEB_H();
	busDataOut((addrL >> 8) & 0x000000FF);
	PIN_NAND_WEB_L();
	PIN_NAND_WEB_H();
	busDataOut((addrL >> 16) & 0x000000FF);
	PIN_NAND_WEB_L();
	PIN_NAND_WEB_H();
	busDataOut((addrL >> 24) & 0x000000FF);
	PIN_NAND_WEB_L();
	PIN_NAND_WEB_H();
	if(MT29F2_4_8G){
		// used for 2Gb and above */
		busDataOut(addrH & 0x000000FF);
		PIN_NAND_WEB_L();
		PIN_NAND_WEB_H();
	}
	PIN_NAND_ALE_L();

	// transfer data to be written (nBytes can be zero)
	for(i = 0; i<nBytes; i++){
		busDataOut(buf[i]);
		PIN_NAND_WEB_L();
		PIN_NAND_WEB_H();
	}
	return NF_OK;
}

// part 2
long nandFlash_program_contd(unsigned long startCol, unsigned long nBytes, char *buf){
	// This function allows one to write (program) additional data in the same page as the one started with nandFlash_page_program_start.
	// This function can be called any number of times in the same page before calling the nandFlash_program_end() closing function.
	// As with all Flash memory, it is illegal to re-write at previously written-to addresses unless the entire block is erased first.
	//unsigned long addrL;
	long i;

	// Even though we just called nandFlash_page_program_start, we still need to select the NAND flash again (and the bus direction)
	// as other FPGA-related transaction(s) may have occurred in-between.
	PIN_NAND_ALE_L();	// ALE = '0'
	PIN_NAND_REB_H();	// RE# = '1'
	NF_FPGAn();	// enable NAND Flash
	busDirOut();

	// We cannot skip the  the busy check. we just called nandFlash_page_program_start or nandFlash_program_contd
	/*
	busDirOut();
	busDataOut(0x85);		// random data input command
	PIN_NAND_CLE_H();
	PIN_NAND_WEB_L();
	PIN_NAND_WEB_H();
	PIN_NAND_CLE_L();

	// enter columns only
	addrL = startCol & 0x0000FFFF;
	// 2 address cycles
	PIN_NAND_ALE_H();
	busDataOut(addrL & 0x000000FF);
	PIN_NAND_WEB_L();
	PIN_NAND_WEB_H();
	busDataOut((addrL >> 8) & 0x000000FF);
	PIN_NAND_WEB_L();
	PIN_NAND_WEB_H();
	PIN_NAND_ALE_L();
	*/
	// transfer data to be written
	for(i = 0; i<nBytes; i++){
		busDataOut(buf[i]);
		PIN_NAND_WEB_L();
		PIN_NAND_WEB_H();
	}
	return NF_OK;
}

// part 3.
long nandFlash_page_program_end(void){
	// USAGE: PAGES MUST BE PROGRAMMED CONSECUTIVELY WITHIN A BLOCK + BLOCK MUST BE ERASED BEFORE RE-PROGRAMMING
	// Write a full page or part thereof, starting at the specified startCol. The number of bytes to be written is nBytes.
	// The data to be written is stored in buf. The maximum number of bytes to be written is 2112 (for a complete page including the ECC redundancy).
	// returns return 0 on success, a negative error number otherwise.
	// see Micron 1Gb NAND flash memory MT29F1Gxx specifications Figure 19
	unsigned char a;

	// trigger programming
	busDataOut(0x10);	// program command
	PIN_NAND_CLE_H();
	PIN_NAND_WEB_L();
	PIN_NAND_WEB_H();
	PIN_NAND_CLE_L();

	// NAND flash starts programming. We must wait for the end of the program period.
	// set timeout awaiting for ready signal from NAND flash
	// wait until NAND flash is no longer busy  tPROG < 700us max
	timer[0] = 2; 	// 10ms units -> 10-20ms
	a = nandFlash_check_status();	// first read status
	while(1){
		if((a & 0x40) == 0x40)
			break;	// ready
		if(!timer[0]){
			PIN_NAND_WPB_L();	// re-instate write protection
			PIN_NAND_CEB_H();	// de-select flash
			return -NF_ETIMEOUT; // timeout
		}
		a = nandFlash_check_status_contd(); // subsequent read status
	}

	PIN_NAND_WPB_L();	// reinstate write protection
	PIN_NAND_CEB_H();	// de-select flash
	if(!(a & 0x01))
		return NF_OK;	// successful erase
	else
		return -NF_EBADBLOCK;	// unsuccessful erase
	return NF_OK;
}


long nandFlash_page_read(unsigned long blockNo, unsigned long pageNo, unsigned long startCol, unsigned long nBytesRead, unsigned char *buf){
	// read a full page or part thereof, starting at the specified startCol. The number of bytes read is nBytesRead.
	// The data read is stored in buf. The maximum number of bytes read is 2112 (for a complete page including the ECC redundancy).
	// returns return 0 on success, a negative error number otherwise.
	// see Micron 1Gb NAND flash memory MT29F1Gxx specifications Figure 12
	unsigned long addrL;
	unsigned long addrH;	// address spans 4(1Gb) or 5(2Gb+) bytes, thus exceeding a long
//	long r_bb, r_bb_prev;
	long i;
	unsigned char a;

	if(blockNo >= NBLOCKS)
		return -NF_EDATAIN;

	PIN_NAND_ALE_L();	// ALE = '0'
	NF_FPGAn();	// enable NAND Flash

	// check busy status before most commands
	timer[0] = 2; 	// 10ms units -> 10-20ms timeout
	a = nandFlash_check_status();	// first read status
	while(1){
		if((a & 0x60) == 0x60)
			break;
		if(!timer[0])
			return -NF_EBUSY;
		a = nandFlash_check_status_contd(); // subsequent read status
	}

	busDirOut();
	busDataOut(0x00);	//page read command
	PIN_NAND_CLE_H();
	PIN_NAND_WEB_L();
	PIN_NAND_WEB_H();
	PIN_NAND_CLE_L();

	// form 5-cycle address array as per Table 4.
	addrH = (blockNo >> 10) & 0x00000003;	// used for 2Gb and above
	addrL = (blockNo << 22) & 0xFFC00000;
	addrL |= (pageNo << 16) & 0x003F0000;
	addrL |= startCol & 0x0000FFFF;

	PIN_NAND_ALE_H();
	busDataOut(addrL & 0x000000FF);
	PIN_NAND_WEB_L();
	PIN_NAND_WEB_H();
	busDataOut((addrL >> 8) & 0x000000FF);
	PIN_NAND_WEB_L();
	PIN_NAND_WEB_H();
	busDataOut((addrL >> 16) & 0x000000FF);
	PIN_NAND_WEB_L();
	PIN_NAND_WEB_H();
	busDataOut((addrL >> 24) & 0x000000FF);
	PIN_NAND_WEB_L();
	PIN_NAND_WEB_H();
	if(MT29F2_4_8G){
		// used for 2Gb and above */
		busDataOut(addrH & 0x000000FF);
		PIN_NAND_WEB_L();
		PIN_NAND_WEB_H();
	}
	PIN_NAND_ALE_L();

	busDataOut(0x30);	//page read command
	PIN_NAND_CLE_H();
	PIN_NAND_WEB_L();
	PIN_NAND_WEB_H();
	PIN_NAND_CLE_L();

	// check R/B# status to know when to read the data we asked for
	timer[0] = 2; 	// 10ms units -> 10-20ms timeout
	a = nandFlash_check_status();	// first read status
	while(1){
		if((a & 0x60) == 0x60)
			break;
		if(!timer[0])
			return -NF_EBUSY;
		a = nandFlash_check_status_contd(); // subsequetn read status
	}

	// now that data is ready, re-issue command 0x00
	busDirOut();
	busDataOut(0x00);	//page read command
	PIN_NAND_CLE_H();
	PIN_NAND_WEB_L();
	PIN_NAND_WEB_H();
	PIN_NAND_CLE_L();
	// then read data
	busDirIn();
	PIN_NAND_REB_H();
	for(i = 0; i<nBytesRead; i++){	// read the requested number of bytes. max is 2112 for a complete page including ECC redundancy.
		PIN_NAND_REB_L();
		PIN_NAND_REB_L();	// delay before reading
		buf[i] = busDataIn();
		PIN_NAND_REB_H();
	}
	return NF_OK;

/*
 * code for hardware rev1
	// NAND flash starts page data transfer. We must wait for the end of the busy period.
	// set timeout awaiting for ready signal from NAND flash
	timer[0] = 100; 	// 10ms units -> 1s timeout
	while(1){
		if(timer[0] == 0){	// timer expired
			return -NF_ETIMEOUT;
		}

		// detect transition from busy to ready
		r_bb_prev = r_bb;  					// previous ready/busy#
		r_bb = nandFlash_check_ready();  	// current ready/busy#
		if(r_bb && (!r_bb_prev))			// transition from busy to ready
			break;
	}

	for(i = 0; i<nBytesRead; i++){	// read the requested number of bytes. max is 2112 for a complete page including ECC redundancy.
		PIN_NAND_REB_L();
		buf[i] = nandFlash_datain();
		PIN_NAND_REB_H();
	}
	PIN_NAND_CEB_H();	// de-select flash
	return NF_OK;
*/
}

long nandFlash_checkBadBlock(unsigned long blockNo){
	// check if a block is marked as bad.
	// Method: a block is marked as bad if the byte at location (page 0, offset 2048 bytes) or (page 1, offset 2048 bytes) is not 0xFF.
	// returns return 0 on success, a negative error number otherwise.
	long retval;
	unsigned char buf[1];

	if(blockNo >= NBLOCKS)
		return -NF_EDATAIN;

	// read the first redundancy byte in page 0 of the specified block
	retval = nandFlash_page_read(blockNo, 0, 2048, 1, &buf[0]);
	if(retval)
		return retval;
	if(buf[0] != 0xFF)
		return -NF_EBADBLOCK;

	// repeat for page 1
	retval = nandFlash_page_read(blockNo, 1, 2048, 1, &buf[0]);
	if(retval)
		return retval;
	if(buf[0] != 0xFF)
		return -NF_EBADBLOCK;

	return NF_OK;
}

long nandFlash_setBadBlock(unsigned long blockNo){
	// Mark a block as 'bad' by writing a non-FF value at the first redundancy byte of the first page.
	// To clear the flag, erase the block (it may fail if the status at the end of the block erase command is bad).
	char buf[1];
	long retval;

	buf[0] = 0x00;
	retval = nandFlash_page_program_start(blockNo, 0, 2048, 1, &buf[0]);
	if(retval)
		return retval;
	retval = nandFlash_page_program_end();
	if(retval)
		return retval;
	return NF_OK;
}

long nandFlash_read(unsigned long start_addr, unsigned char *buf, unsigned long nBytes){
	// copy nBytes data bytes from the NAND flash to the buffer buf[], starting at the specified flash address.
	// Invisible to the user is the bad block management. Read accesses to bad blocks are re-directed to good blocks
	// at the upper end of the NAND flash memory map.
	// Pre-requisite: CE# must be enabled prior
	unsigned long startBlock, startPage, startColumn;
	unsigned long endBlock, endPage, endColumn;
	unsigned long iblock, ipage, icol;
	unsigned long iblock_alt;	// alternative block number if nominal block is flagged as 'bad'
	unsigned long end_addr;
	unsigned long wptr, nBytesReq;
	long retval;

	// convert start/end memory addresses to start/end (block,page,column) coordinates
	startBlock = start_addr / BLOCKSIZE;
	startPage = (start_addr % BLOCKSIZE) / PAGESIZE;
	startColumn = start_addr % 2048;

	end_addr = start_addr + nBytes;
	endBlock = end_addr / BLOCKSIZE;
	endPage = (end_addr % BLOCKSIZE) / PAGESIZE;
	endColumn = end_addr % PAGESIZE;

	wptr = 0;	// initialize the write pointer

	// read one page at a time
	iblock = startBlock;
	ipage = startPage;
	icol = startColumn;
	for(; iblock <= endBlock; iblock++){
		// redirect if bad block
		iblock_alt = nfInfo.blockRedirect[iblock];
		// 64 pages per block
		for(; ipage < 64; ipage++){
			// read up to 2KB page from the NAND flash memory
			nBytesReq = 2048 - icol;	// maximum number of bytes in current page, starting at icol.
			// make sure we don't read more than the output buffer can hold.
			if(nBytesReq > (nBytes-wptr))
				nBytesReq = nBytes-wptr;
			retval = nandFlash_page_read(iblock_alt, ipage, icol, nBytesReq, &buf[wptr]);
			if(retval)
				return retval;
			wptr += nBytesReq;
			// rewind icolumn
			icol = 0;
		}
		// rewind ipage
		ipage = 0;
	}

	return NF_OK;
}

//--------------------------------------------------------------------------------------
// higher level user functions (include automatic (i.e. invisible to the user) bad sector management)

long nandFlash_EFS(unsigned long blockNo){
	// Erase flash sector (old terminology: this is really a 128KB block in NAND flash block)
	// redirect based on the bad block table
	long retval;
	unsigned long phyBlockNo;

	if(blockNo >= (NBLOCKS - nfInfo.nBadBlocks))
		return -NF_EDATAIN;

	// re-direct to a good block
	phyBlockNo = nfInfo.blockRedirect[blockNo];

	retval = nandFlash_blockErase(phyBlockNo);
	if(retval == NF_EBADBLOCK){
		// bad sector.. what to do???
	}
	return retval;
}

long nandFlash_GCS(unsigned long blockNo, unsigned long *checksum){
	// GCS = get checksum
	// Compute and return the 16-bit checksum for the specified NAND flash block
	// (data only, excluding the 4KB FEC redundancy field)
	// Invisible to the user is the bad block management. Read accesses to bad blocks are re-directed to good blocks
	// at the upper end of the NAND flash memory map.
	long retval;
	unsigned long phyBlockNo;
	unsigned long ipageno, i;
	unsigned char buf[2048];

	// first check block number is valid
	if(blockNo >= (NBLOCKS - nfInfo.nBadBlocks))
		return -NF_EDATAIN;	// block number out of range

	// re-direct to a good block
	phyBlockNo = nfInfo.blockRedirect[blockNo];

	*checksum = 0;

	PIN_NAND_ALE_L();	// ALE = '0'
	NF_FPGAn();	// enable NAND Flash

	for(ipageno = 0; ipageno < 64; ipageno++){
		retval = nandFlash_page_read(phyBlockNo, ipageno, 0, sizeof(buf), buf);
		for(i = 0; i<sizeof(buf); i++){
			*checksum += ((unsigned long)(buf[i])) & 0x000000FF;
		}
		if(retval)
			return retval;
	}
	// truncate to 16-bit checksum
	*checksum &= 0x0000FFFF;

	PIN_NAND_CEB_H();	// User transaction is complete -> release NAND flash chip select

	return NF_OK;

}
